"""
@file
    obsw_tm_listener.py
@date
    01.11.2019
@brief
    Separate class to listen to telecommands.
    This will enable to run the listener in a separate thread later.
    This Listener will propably have some kind of mode. In default configuration,
    it will just listen for packets
"""
import time
import threading
from collections import deque
from typing import TypeVar

from comIF.obsw_com_interface import ComIfT
import config.OBSW_Config as g

TmListenerT = TypeVar('TmListenerT', bound='TmListener')


class TmListener:
    """
    Performs all TM listening operations.
    This listener can be used by setting the modeChangeEvent Event with the set() function
    and changing the mode to do special mode operations. The mode operation ends as soon
    the modeOpFinished Event is set() !
    """
    def __init__(self, comInterface: ComIfT, tmTimeout: float, tcTimeoutFactor: float):
        self.tmTimeout = tmTimeout
        self.tcTimeoutFactor = tcTimeoutFactor
        self.comInterface = comInterface
        # this will be the default mode (listener mode)
        self.modeId = g.ModeList.ListenerMode
        # TM Listener operations can be suspended by setting this flag
        self.listenerActive = threading.Event()
        self.listenerActive.set()
        # I don't think a listener is useful without the main program,
        # so we might just declare it daemonic.
        # UPDATE: Right now, the main program is not in a permanent loop and setting the
        # thread daemonic will cancel the program.
        # Solved for now by setting a permanent loop at the end of the main program
        self.listenerThread = threading.Thread(target=self.performOperation, daemon=True)
        # This Event is set by sender objects to perform mode operations
        self.modeChangeEvent = threading.Event()
        # This Event is set by sender objects if all necessary operations are done
        # to transition back to listener mode
        self.modeOpFinished = threading.Event()
        # maybe we will just make the thread daemonic...
        # self.terminationEvent = threading.Event()
        # This Event is set and cleared by the listener to inform the sender objects
        # if a reply has been received
        self.replyEvent = threading.Event()
        # Will be filled for the Unit Test
        self.tmInfoQueue = deque()

    def start(self):
        self.listenerThread.start()

    def performOperation(self):
        while True:
            if self.listenerActive.is_set():
                self.defaultOperation()
            else:
                time.sleep(1)

    def defaultOperation(self):
        self.comInterface.poll_interface()
        if self.modeChangeEvent.is_set():
            # TODO: We should put this in a timeout.. Each mode operation up until now only takes
            #       a maximum specified time (software test 5 minutes maybe?).
            #       Otherwise, this is a permanent loop
            self.modeChangeEvent.clear()
            while not self.modeOpFinished.is_set():
                self.performModeOperation()
            self.modeOpFinished.clear()
            print("Transitioning to listener mode.")
            self.modeId = g.ModeList.ListenerMode

    def performModeOperation(self):
        """
        By setting the modeChangeEvent with set() and specifying the mode variable,
        the TmListener is instructed to perform certain operations.
        :return:
        """
        # Listener Mode
        if self.modeId == g.ModeList.ListenerMode:
            pass
        # Single Command Mode
        elif self.modeId == g.ModeList.SingleCommandMode:
            # Listen for one reply sequence.
            if self.checkForOneTelemetrySequence():
                # Set reply event, will be cleared by checkForFirstReply()
                self.replyEvent.set()
        # Sequential Command Mode
        elif self.modeId == g.ModeList.ServiceTestMode or \
                self.modeId == g.ModeList.SoftwareTestMode:
            if self.checkForOneTelemetrySequence():
                print("TM Listener: Reply sequence received!")
                self.replyEvent.set()
        elif self.modeId == g.ModeList.UnitTest:
            self.tmInfoQueue = self.comInterface.receive_telemetry_and_store_info(self.tmInfoQueue)
            self.replyEvent.set()

    def checkForOneTelemetrySequence(self) -> bool:
        """
        Receive all telemetry for a specified time period.
        This function prints the telemetry sequence but does not return it.
        :return:
        """
        tmReady = self.comInterface.data_available(self.tmTimeout * self.tcTimeoutFactor)
        if tmReady is False:
            return False
        else:
            self.comInterface.receive_telemetry()
            start_time = time.time()
            elapsed_time = 0
            while elapsed_time < self.tmTimeout:
                tmReady = self.comInterface.data_available(1.0)
                if tmReady:
                    self.comInterface.receive_telemetry()
                elapsed_time = time.time() - start_time
            # the timeout value can be set by special TC queue entries if packet handling
            # takes longer, but it is reset here to the global value
            if self.tmTimeout is not g.G_TM_TIMEOUT:
                self.tmTimeout = g.G_TM_TIMEOUT
            return True
