# -*- coding: utf-8 -*-
"""
Created on Wed Apr  4 11:43:00 2018

@author: S. Gaisser, R. Mueller
"""

import crcmod
import logging
from typing import TypeVar, Dict, Union, Tuple, Deque

PusTcT = TypeVar('PusTcT', bound='PUSTelecommand')
PusTcInfoT = Union[Dict[str, any], None]
PusTcTupleT = Tuple[bytearray, PusTcInfoT]
TcAuxiliaryTupleT = Tuple[str, any]
# Union: Type can be either of those 2
TcQueueEntryT = Union[TcAuxiliaryTupleT, PusTcTupleT]
TcQueueT = Deque[TcQueueEntryT]
PusTcInfoQueueT = Deque[PusTcInfoT]


class PusTelecommand:
    headerSize = 6

    def __init__(self, service: int, subservice: int, SSC=0, data=bytearray([]), sourceId: int = 0,
                 version: int = 0, packetType: int = 1, dataFieldHeaderFlag: int = 1,
                 apid: int = 0x73, length: int = 0):
        self.packet_id = [0x0, 0x0]
        self.packet_id[0] = ((version << 5) & 0xE0) | ((packetType & 0x01) << 4) | \
                            ((dataFieldHeaderFlag & 0x01) << 3) | ((apid & 0x700) >> 8)
        self.packet_id[1] = apid & 0xFF
        self.ssc = SSC
        self.psc = (SSC & 0x3FFF) | (0xC0 << 8)
        self.length = length
        self.pusVersionAndAckByte = 0b00011111
        self.service = service
        self.subservice = subservice
        self.sourceId = sourceId
        self.data = data

    def getLength(self) -> int:
        try:
            length = 4 + len(self.data) + 1
            return length
        except TypeError:
            print("OBSW_TcPacket: Invalid service_type of data")
            logging.exception("Error")
            exit()

    def pack(self) -> bytearray:
        dataToPack = bytearray()
        dataToPack.append(self.packet_id[0])
        dataToPack.append(self.packet_id[1])
        dataToPack.append((self.psc & 0xFF00) >> 8)
        dataToPack.append(self.psc & 0xFF)
        length = self.getLength()
        dataToPack.append((length & 0xFF00) >> 8)
        dataToPack.append(length & 0xFF)
        dataToPack.append(self.pusVersionAndAckByte)
        dataToPack.append(self.service)
        dataToPack.append(self.subservice)
        dataToPack.append(self.sourceId)
        dataToPack += self.data
        crc_func = crcmod.mkCrcFun(0x11021, rev=False, initCrc=0xFFFF, xorOut=0x0000)
        crc = crc_func(dataToPack)

        dataToPack.append((crc & 0xFF00) >> 8)
        dataToPack.append(crc & 0xFF)
        return dataToPack

    def packInformation(self) -> PusTcInfoT:
        tcInformation = {
            "service": self.service,
            "subservice": self.subservice,
            "ssc": self.ssc,
            "packet_id": self.packet_id,
            "data": self.data
        }
        return tcInformation

    def packCommandTuple(self) -> PusTcTupleT:
        commandTuple = (self.pack(), self.packInformation())
        return commandTuple

    def print(self):
        packet = self.pack()
        print("Command in Hexadecimal: [", end="")
        for counter in range(len(packet)):
            if counter == len(packet) - 1:
                print(str(hex(packet[counter])), end="")
            else:
                print(str(hex(packet[counter])) + ", ", end="")
        print("]")


# Takes pusPackets, removes current Packet Error Control,
# calculates new CRC (16 bits at packet end) and
# adds it as correct Packet Error Control Code. Reference: ECSS-E70-41A p. 207-212
def generatePacketCRC(TCPacket: PusTcT) -> PusTcT:
    crc_func = crcmod.mkCrcFun(0x11021, rev=False, initCrc=0xFFFF, xorOut=0x0000)
    crc = crc_func(bytearray(TCPacket[0:len(TCPacket) - 2]))
    TCPacket[len(TCPacket) - 2] = (crc & 0xFF00) >> 8
    TCPacket[len(TCPacket) - 1] = crc & 0xFF
    return TCPacket


def generateCRC(data: bytearray) -> bytearray:
    dataWithCRC = bytearray()
    dataWithCRC += data
    crc_func = crcmod.mkCrcFun(0x11021, rev=False, initCrc=0xFFFF, xorOut=0x0000)
    crc = crc_func(data)
    dataWithCRC.append((crc & 0xFF00) >> 8)
    dataWithCRC.append(crc & 0xFF)
    return dataWithCRC

# pylint: disable=line-too-long

# Structure of a PUS TC Packet :
# A PUS packet consists of consecutive bits, the allocation and structure is standardised.
# Extended information can be found in ECSS-E-70-41A  on p.42
# The easiest form to send a PUS Packet is in hexadecimal form.
# A two digit hexadecimal number equals one byte, 8 bits or one octet
# o = optional, Srv = Service
#
# The structure is shown as follows for TC[17,1]
# 1. Structure Header
# 2. Structure Subheader
# 3. Component (size in bits)
# 4. Hexadecimal number
# 5. Binary Number
# 6. Decimal Number
#
# -------------------------------------------Packet Header(48)------------------------------------------|   Packet   |
#  ----------------Packet ID(16)----------------------|Packet Sequence Control (16)| Packet Length (16) | Data Field |
# Version       | Type(1) |Data Field    |APID(11)    | SequenceFlags(2) |Sequence |                    | (Variable) |
# Number(3)     |         |Header Flag(1)|            |                  |Count(14)|                    |            |
#           0x18               |    0x73              |       0xc0       | 0x19    |   0x00  |   0x04   |            |
#    000      1      1      000|  01110011            | 11  000000       | 00011001|00000000 | 0000100  |            |
#     0   |   1   |  1     |    115(ASCII s)          | 3 |            25          |   0     |    4     |            |
#
#   - Packet Length is an unsigned integer C = Number of Octets in Packet Data Field - 1
#
# Packet Data Field Structure:
#
# ------------------------------------------------Packet Data Field------------------------------------------------- |
# ---------------------------------Data Field Header ---------------------------|AppData|Spare|    PacketErrCtr      |
# CCSDS(1)|TC PUS Ver.(3)|Ack(4)|SrvType (8)|SrvSubtype(8)|Source ID(o)|Spare(o)|  (var)|(var)|         (16)         |
#        0x11 (0x1F)            |  0x11     |   0x01      |            |        |       |     | 0xA0     |    0xB8   |
#    0     001     1111         |00010001   | 00000001    |            |        |       |     |          |           |
#    0      1       1111        |    17     |     1       |            |        |       |     |          |           |
#
#   - The source ID is present as one byte. For now, ground = 0x00.
