"""
@brief:     Generic PUS packet class to deserialize raw PUS telemetry.
@date:      09.04.2020
@author:    R.Mueller, S. Gaisser
"""
import datetime
from typing import TypeVar, Dict, Tuple, Deque, List, Final

from crcmod import crcmod


PusTmT = TypeVar('PusTmT', bound='PUSTelemetry')
PusTmInfoT = Dict[str, any]
PusTmTupleT = Tuple[PusTmInfoT, PusTmT]

PusTmQueueT = Deque[PusTmT]
PusTmListT = List[PusTmT]
PusTmInfoQueueT = Deque[PusTmInfoT]
PusTmTupleQueueT = Deque[PusTmTupleT]


class PusTelemetry:
    """
    Generic PUS telemetry class. It is instantiated by passing the raw pus telemetry packet
    (bytearray) to the constructor. It automatically deserializes the packet, exposing
    various packet fields via getter functions.
    """
    def __init__(self, byte_array: bytes):
        self._packet_raw: Final = byte_array
        self._pus_header: Final = PusPacketHeader(byte_array)
        byte_array = byte_array[6:]
        self._data_field_header: Final = ObswPusPacketDataFieldHeader(byte_array)
        byte_array = byte_array[12:]
        self._tm_data: Final = byte_array[:len(byte_array) - 2]
        self._crc: Final = byte_array[len(byte_array) - 2] << 8 | byte_array[len(byte_array) - 1]
        self.print_info = ""

    def get_service(self):
        """
        :return: Service ID
        """
        return self._data_field_header.service_type

    def get_subservice(self):
        """
        :return: Subservice ID
        """
        return self._data_field_header.service_subtype

    def get_tm_data(self) -> bytearray:
        """
        :return: TM application data (raw)
        """
        return self._tm_data

    def pack_tm_information(self) -> PusTmInfoT:
        """
        Packs important TM information needed for tests in a convenient dictionary
        :return: TM dictionary
        """
        tm_information = {
            "service": self.get_service(),
            "subservice": self.get_subservice(),
            "ssc": self.get_ssc(),
            "data": self._tm_data,
            "crc": self._crc,
            "valid": self._pus_header.valid
        }
        return tm_information

    def specify_packet_info(self, print_info: str):
        """
        Caches a print information string for later printing
        :param print_info:
        :return:
        """
        self.print_info = print_info

    def append_packet_info(self, print_info: str):
        """
        Similar to the function above, but appends to the existing information string.
        :param print_info:
        :return:
        """
        self.print_info = self.print_info + print_info

    def append_telemetry_content(self, content_list: list):
        """
        Default implementation adds the PUS header content to the list which can then be
        printed with a simple print() command. To add additional content, override this method
        (don't forget to still call this function with super() if the header is required)
        :param content_list: Header content will be appended to this list
        :return:
        """
        self._data_field_header.append_data_field_header(content_list)
        self._pus_header.append_pus_packet_header(content_list)

    def append_telemetry_column_headers(self, header_list: list):
        """
        Default implementation adds the PUS header content header (confusing, I know)
        to the list which can then be  printed with a simple print() command.
        To add additional headers, override this method
        (don't forget to still call this function with super() if the header is required)
        :param header_list: Header content will be appended to this list
        :return:
        """
        self._data_field_header.append_data_field_header_column_header(header_list)
        self._pus_header.append_pus_packet_header_column_headers(header_list)

    def get_raw_packet(self) -> bytearray:
        """
        Get the whole TM packet as a bytearray (raw)
        :return: TM packet
        """
        return self._packet_raw

    def get_packet_size(self) -> int:
        """
        :return: Size of the TM packet
        """
        # PusHeader Size + _tm_data size
        size = PusPacketHeader.PUS_HEADER_SIZE + self._pus_header.length + 1
        return size

    def get_ssc(self) -> int:
        """
        Get the source sequence count
        :return: Source Sequence Count (see below, or PUS documentation)
        """
        return self._pus_header.source_sequence_count

    def return_data_string(self) -> str:
        """
        Returns the TM data in a clean printable string format
        :return:
        """
        str_to_print = "["
        for byte in self._tm_data:
            str_to_print += str(hex(byte)) + " , "
        str_to_print = str_to_print.rstrip(',', )
        str_to_print += ']'
        return str_to_print

    def print_data(self):
        """
        Prints the TM data in a clean format
        :return:
        """
        print(self.return_data_string())


# pylint: disable=too-many-instance-attributes
class PusPacketHeader:
    """
    This class unnpacks the PUS packet header (see PUS structure below or PUS documentation)
    """
    PUS_HEADER_SIZE = 6

    def __init__(self, pus_packet_raw: bytes):
        data_to_check = pus_packet_raw
        if len(pus_packet_raw) < PusPacketHeader.PUS_HEADER_SIZE:
            self.version = 0
            self.type = 0
            self.data_field_header_flag = 0
            self.apid = 0
            self.segmentation_flag = 0
            self.source_sequence_count = 0
            self.length = 0
            self.valid = False
            return
        self.version = pus_packet_raw[0] >> 5
        self.type = pus_packet_raw[0] & 0x10
        self.data_field_header_flag = (pus_packet_raw[0] & 0x8) >> 3
        self.apid = ((pus_packet_raw[0] & 0x7) << 8) | pus_packet_raw[1]
        self.segmentation_flag = (pus_packet_raw[2] & 0xC0) >> 6
        self.source_sequence_count = ((pus_packet_raw[2] & 0x3F) << 8) | pus_packet_raw[3]
        self.length = pus_packet_raw[4] << 8 | pus_packet_raw[5]
        self.valid = False
        crc_func = crcmod.mkCrcFun(0x11021, rev=False, initCrc=0xFFFF, xorOut=0x0000)
        if len(data_to_check) < ((self.length + 1) + PusPacketHeader.PUS_HEADER_SIZE):
            print("Invalid packet length")
            return
        data_to_check = data_to_check[0:(self.length + 1) + PusPacketHeader.PUS_HEADER_SIZE]
        crc = crc_func(data_to_check)
        if crc == 0:
            self.valid = True
        else:
            print("Invalid CRC detected !")

    def append_pus_packet_header(self, array):
        array.append(str(chr(self.apid)))
        array.append(str(self.source_sequence_count))
        if self.valid:
            array.append("Yes")
        else:
            array.append("No")
        # header_list.append(str(self.valid))

    @staticmethod
    def append_pus_packet_header_column_headers(array):
        array.append("APID")
        array.append("SSC")
        array.append("Packet Valid")


class ObswPusPacketDataFieldHeader:
    def __init__(self, bytesArray):
        self.pus_version_and_ack_byte = (bytesArray[0] & 0x70) >> 4
        self.service_type = bytesArray[1]
        self.service_subtype = bytesArray[2]
        self.subcounter = bytesArray[3]
        self.time = ObswTimestamp(bytesArray[4:13])

    def append_data_field_header(self, content_list: list):
        """
        Append important data field header parameters to the passed content list.
        :param content_list:
        :return:
        """
        content_list.append(str(self.service_type))
        content_list.append(str(self.service_subtype))
        content_list.append(str(self.subcounter))
        self.time.print_time(content_list)

    def append_data_field_header_column_header(self, header_list: list):
        """
        Append important data field header column headers to the passed list.
        :param header_list:
        :return:
        """
        header_list.append("Service")
        header_list.append("Subservice")
        header_list.append("Subcounter")
        self.time.print_time_headers(header_list)


class ObswTimestamp:
    """
    Unpacks the time datafield of the TM packet.
    """
    def __init__(self, byteArray):
        # pField = byte_array[0]
        byteArray = byteArray[1:]
        self.days = ((byteArray[0] << 8) | (byteArray[1])) - 4383
        self.seconds = self.days * (24 * 60 * 60)
        s_day = ((byteArray[2] << 24) | (byteArray[3] << 16) |
                 (byteArray[4]) << 8 | byteArray[5]) / 1000
        self.seconds += s_day
        self.time = self.seconds
        self.datetime = str(datetime.datetime.
                            utcfromtimestamp(self.time).strftime("%Y-%m-%d %H:%M:%S.%f"))

    def print_time(self, array):
        array.append(self.time)
        array.append(self.datetime)

    @staticmethod
    def print_time_headers(array):
        array.append("OBSWTime (s)")
        array.append("Time")

# pylint: disable=line-too-long
# Structure of a PUS Packet :
# A PUS packet consists of consecutive bits, the allocation and structure is standardised.
# Extended information can be found in ECSS-E-70-41A  on p.42
# The easiest form to send a PUS Packet is in hexadecimal form.
# A two digit hexadecimal number equals one byte, 8 bits or one octet
# o = optional, Srv = Service
#
# The structure is shown as follows for TM[17,2]
# 1. Structure Header
# 2. Structure Subheader
# 3. Component (size in bits)
# 4. Hexadecimal number
# 5. Binary Number
# 6. Decimal Number
#
# -------------------------------------------Packet Header(48)------------------------------------------|   Packet   |
#  ----------------Packet ID(16)----------------------|Packet Sequence Control (16)| Packet Length (16) | Data Field |
# Version       | Type(1) |Data Field    |APID(11)    | SequenceFlags(2) |Sequence |                    | (Variable) |
# Number(3)     |         |Header Flag(1)|            |                  |Count(14)|                    |            |
#           0x18               |    0x73              |       0xc0       | 0x19    |   0x00  |   0x04   |            |
#    000      1      0      000|  01110011            | 11  000000       | 00011001|00000000 | 0000100  |            |
#     0   |   1   |  0     |    115(ASCII s)          | 3 |            25          |   0     |    4     |            |
#
#   - Packet Length is an unsigned integer C = Number of Octets in Packet Data Field - 1
#
# Packet Data Field Structure:
#
# ------------------------------------------------Packet Data Field------------------------------------------------- |
# ---------------------------------Data Field Header ---------------------------|AppData|Spare|    PacketErrCtr      |
# CCSDS(1)|TC PUS Ver.(3)|Ack(4)|SrvType (8)|SrvSubtype(8)|Source ID(o)|Spare(o)|  (var)|(var)|         (16)         |
#        0x11 (0x1F)            |  0x11     |   0x01      |            |        |       |     | Calc.    |    Calc.  |
#    0     001     1111         |00010001   | 00000001    |            |        |       |     |          |           |
#    0      1       1111        |    17     |     2       |            |        |       |     |          |           |
#
#   - The source ID is present as one byte. Is it necessary? For now, ground = 0x00.
